from __future__ import annotations

from sarus.dataspec_wrapper import DataSpecWrapper
from sarus.utils import register_ops, sarus_init

try:
    import imblearn.over_sampling as imb_over_sampling
except ModuleNotFoundError:
    pass  # error message in sarus_data_spec.typing


class SMOTENC(DataSpecWrapper[imb_over_sampling.SMOTENC]):
    @sarus_init("imblearn.IMB_SMOTENC")
    def __init__(
        categorical_features,
        *,
        sampling_strategy="auto",
        random_state=None,
        k_neighbors=5,
        n_jobs=None,
    ):
        ...


register_ops()



import numpy as np

from sarus.dataspec_wrapper import DataSpecWrapper


class Bool(DataSpecWrapper[np.bool_]):
    ...


class Byte(DataSpecWrapper[np.byte]):
    ...


class UByte(DataSpecWrapper[np.ubyte]):
    ...


class Short(DataSpecWrapper[np.short]):
    ...


class UShort(DataSpecWrapper[np.ushort]):
    ...


class Intc(DataSpecWrapper[np.intc]):
    ...


class UIntc(DataSpecWrapper[np.uintc]):
    ...


class Int_(DataSpecWrapper[np.int_]):
    ...


class UInt_(DataSpecWrapper[np.uint]):
    ...


class LongLong(DataSpecWrapper[np.longlong]):
    ...


class ULongLong(DataSpecWrapper[np.ulonglong]):
    ...


class Single(DataSpecWrapper[np.single]):
    ...


class Double(DataSpecWrapper[np.double]):
    ...


class LongDouble(DataSpecWrapper[np.longdouble]):
    ...


class CSingle(DataSpecWrapper[np.csingle]):
    ...


class CDouble(DataSpecWrapper[np.cdouble]):
    ...


class CLongDouble(DataSpecWrapper[np.clongdouble]):
    ...



from typing import Any, Dict, Type, Union, cast

import sarus_data_spec.protobuf as sp
import sarus_data_spec.typing as st
from sarus_data_spec.context import global_context

from sarus.dataspec_wrapper import DataSpecWrapper
from sarus.manager.typing import SDKManager
from sarus.utils import convert_container


class DataSpecWrapperFactory:
    def __init__(self):
        self._registry = {}

    def register(
        self,
        python_classname: str,
        sarus_wrapper_class: Type[DataSpecWrapper],
    ) -> None:
        self._registry[python_classname] = sarus_wrapper_class

    def create(self, dataspec: st.DataSpec) -> Union[DataSpecWrapper, Any]:
        """Create a wrapper class from a DataSpec.

        If the dataspec's python value is not managed by the SDK, returns an
        unwrapped Python object.
        """
        context = global_context()
        manager: SDKManager = context.manager()
        python_type = manager.python_type(dataspec)
        SarusWrapperClass = self._registry.get(python_type)
        if SarusWrapperClass:
            return SarusWrapperClass.from_dataspec(dataspec)
        else:
            # Datasets are either pd.DataFrame or pd.Series
            if dataspec.prototype() != sp.Scalar:
                raise TypeError(f"Expected a Scalar {dataspec}")
            scalar = cast(st.Scalar, dataspec)
            alt_scalar = scalar.variant(
                kind=st.ConstraintKind.SYNTHETIC, public_context=[]
            )
            return alt_scalar.value()

    def from_value(self, value: Any) -> DataSpecWrapper:
        """Creates a wrapper class from a Python object.

        The Python object can be a container, in which wase it might contain a
        DataSpecWrapper nested deeper. Such containers will be converted to
        DataSpecWrappers.
        """
        value = convert_container(value)
        if isinstance(value, DataSpecWrapper):
            return value

        python_type = str(type(value))
        SarusWrapperClass = self.registry().get(python_type)
        if SarusWrapperClass is None:
            raise TypeError(
                f"Cannot convert type {python_type} to DataSpecWrapper."
            )

        return SarusWrapperClass.from_value(value)

    def registry(self) -> Dict[str, Type[DataSpecWrapper]]:
        return self._registry



from __future__ import annotations

import logging
import typing as t

import pandas as pd

from sarus.dataspec_wrapper import (
    IGNORE_WARNING,
    DataSpecVariant,
    DataSpecWrapper,
)
from sarus.typing import SPECIAL_WRAPPER_ATTRIBUTES
from sarus.utils import (
    create_lambda_op,
    create_method,
    register_ops,
    sarus_init,
    sarus_method,
    sarus_property,
)
from shap.models import Model
from shap.maskers import Masker
import shap
from sarus.dataspec_wrapper import DataSpecWrapper
from shap.explainers import GPUTree, Linear, Tree
from shap import KernelExplainer, Explainer
from sarus.utils import register_ops, sarus_method

import shap
from typing import Any, Union, Optional, Tuple
from sklearn.base import BaseEstimator
import numpy as np
import pandas as pd
from scipy.sparse import spmatrix


class Explainer(DataSpecWrapper[Explainer]):
    @sarus_init("shap.SHAP_EXPLAINER")
    def __init__(
        self,
        model: Union[Callable, BaseEstimator],
        masker: Optional[Union[Callable, ndarray, DataFrame]] = None,
        link: Optional[Callable] = None,
        algorithm: str = "auto",
        output_names: Optional[List[str]] = None,
        feature_names: Optional[List[str]] = None,
        linearize_link: bool = True,
        seed: Optional[int] = None,
    ):
        ...

    @sarus_method("shap.SHAP_SAVE")
    def save(
        self,
        out_file: Any,
        model_saver: str = ".save",
        masker_saver: str = ".save",
    ):
        ...

    @classmethod
    @sarus_method("shap.SHAP_LOAD")
    def load(
        cls,
        in_file: Any,
        model_loader: Callable = Model.load,
        masker_loader: Callable = Masker.load,
        instantiate: bool = True,
    ):
        ...

    @sarus_method("shap.SHAP_EXPLAIN_ROW")
    def explain_row(
        self,
        row_args: Any = None,
        max_evals: Any = None,
        main_effects: Any = None,
        error_bounds: Any = None,
        batch_size: Any = None,
        outputs: Any = None,
        silent: bool = None,
    ):
        ...

    @sarus_method("shap.SHAP_SHAP_VALUES")
    def shap_values(self, X: Union[ndarray, DataFrame]):
        ...

    @sarus_method("shap.SHAP_CALL")
    def __call__(self, X: Union[ndarray, DataFrame]):
        ...


class Tree(DataSpecWrapper[shap.explainers.Tree]):
    @sarus_init("shap.SHAP_TREE")
    def __init__(
        self,
        model: Any,
        data: Optional[Union[ndarray, pd.DataFrame]] = None,
        model_output: str = 'raw',
        feature_perturbation: str = 'interventional',
        feature_names: Optional[List[str]] = None,
        approximate: bool = False,
    ):
        ...

    @sarus_method("shap.SHAP_TREE_SHAP_VALUES")
    def shap_values(
        self,
        X: Union[np.ndarray, pd.DataFrame],
        y: Optional[np.ndarray] = None,
        tree_limit: Optional[int] = None,
        approximate: bool = False,
        check_additivity: bool = True,
        from_call: bool = False,
    ):
        ...

    @sarus_method("shap.SHAP_TREE_SHAP_INTERACTION_VALUES")
    def shap_interaction_values(
        self,
        X: Union[np.ndarray, pd.DataFrame],
        y: Optional[np.ndarray] = None,
        tree_limit: Optional[int] = None,
    ):
        ...

    @sarus_method("shap.SHAP_EXPLAIN_ROW")
    def explain_row(
        self,
        row_args: Any = None,
        max_evals: Any = None,
        main_effects: Any = None,
        error_bounds: Any = None,
        batch_size: Any = None,
        outputs: Any = None,
        silent: bool = None,
    ):
        ...

    @sarus_method("shap.SHAP_CALL")
    def __call__(self, X: Union[ndarray, DataFrame]):
        ...


class GPUTree(DataSpecWrapper[shap.explainers.GPUTree]):
    @sarus_init("shap.SHAP_GPUTREE")
    def __init__(
        self,
        model: Any,
        data: Optional[Union[np.ndarray, pd.DataFrame]],
        model_output: str = 'raw',
        feature_perturbation: str = 'interventional',
        feature_names: Optional[List[str]] = None,
        approximate: bool = False,
    ):
        ...

    @sarus_method("shap.SHAP_TREE_SHAP_VALUES")
    def shap_values(
        self,
        X: Union[np.ndarray, pd.DataFrame],
        y: Optional[np.ndarray] = None,
        tree_limit: Optional[int] = None,
        approximate: bool = False,
        check_additivity: bool = True,
        from_call: bool = False,
    ):
        ...

    @sarus_method("shap.SHAP_TREE_SHAP_INTERACTION_VALUES")
    def shap_interaction_values(
        self,
        X: Union[np.ndarray, pd.DataFrame],
        y: Optional[np.ndarray] = None,
        tree_limit: Optional[int] = None,
    ):
        ...

    @sarus_method("shap.SHAP_EXPLAIN_ROW")
    def explain_row(
        self,
        row_args: Any = None,
        max_evals: Any = None,
        main_effects: Any = None,
        error_bounds: Any = None,
        batch_size: Any = None,
        outputs: Any = None,
        silent: bool = None,
    ):
        ...

    @sarus_method("shap.SHAP_CALL")
    def __call__(self, X: Union[ndarray, DataFrame]):
        ...


class Kernel(DataSpecWrapper[shap.KernelExplainer]):
    @sarus_init("shap.SHAP_KERNEL")
    def __init__(
        self,
        model: Callable,
        data: Union[np.ndarray, pd.DataFrame, spmatrix],
        link: Any = None,
    ):
        ...

    @sarus_method("shap.SHAP_RUN")
    def run(self):
        ...

    @sarus_method("shap.SHAP_ALLOCATE")
    def allocate(self):
        ...

    @sarus_method("shap.SHAP_SOLVE")
    def solve(self, fraction_evaluated: Any, dim: Any):
        ...

    @sarus_method("shap.SHAP_VARYING_GROUPS")
    def varying_groups(self, X: Any):
        ...

    @sarus_method("shap.SHAP_EXPLAIN")
    def explain(self, incoming_instance: Any):
        ...

    @sarus_method("shap.SHAP_ADD_SAMPLE")
    def addsample(self, x: np.array, m: np.array, w: float):
        ...

    @sarus_method("shap.SHAP_EXPLAIN_ROW")
    def explain_row(
        self,
        row_args: Any = None,
        max_evals: Any = None,
        main_effects: Any = None,
        error_bounds: Any = None,
        batch_size: Any = None,
        outputs: Any = None,
        silent: bool = None,
    ):
        ...

    @sarus_method("shap.SHAP_SHAP_VALUES")
    def shap_values(self, X: Union[ndarray, DataFrame]):
        ...

    @sarus_method("shap.SHAP_CALL")
    def __call__(self, X: Union[ndarray, DataFrame]):
        ...


class Linear(DataSpecWrapper[shap.explainers.Linear]):
    @sarus_init("shap.SHAP_LINEAR")
    def __init__(
        self,
        model: Union[BaseEstimator, Tuple[Any, Any]],
        masker: Union[Tuple[Any, Any], np.ndarray, pd.DataFrame, spmatrix],
        link: Any = None,
        nsamples: int = 1000,
        feature_perturbation: Optional[str] = None,
    ):
        ...

    @sarus_method("shap.SHAP_EXPLAIN_ROW")
    def explain_row(
        self,
        row_args: Any = None,
        max_evals: Any = None,
        main_effects: Any = None,
        error_bounds: Any = None,
        batch_size: Any = None,
        outputs: Any = None,
        silent: bool = None,
    ):
        ...

    @sarus_method("shap.SHAP_SHAP_VALUES")
    def shap_values(self, X: Union[ndarray, DataFrame]):
        ...

    @sarus_method("shap.SHAP_CALL")
    def __call__(self, X: Union[ndarray, DataFrame]):
        ...


class Partition(DataSpecWrapper[shap.explainers.Partition]):
    @sarus_init("shap.SHAP_PARTITION")
    def __init__(
        self,
        model: Union[BaseEstimator, Tuple[Any, Any]],
        masker: Union[Tuple[Any, Any], np.ndarray, pd.DataFrame, spmatrix],
        output_names: Any = None,
        link: Any = None,
        nsamples: int = 1000,
        feature_perturbation: Optional[str] = None,
    ):
        ...

    @sarus_method("shap.SHAP_EXPLAIN_ROW")
    def explain_row(
        self,
        row_args: Any = None,
        max_evals: Any = None,
        main_effects: Any = None,
        error_bounds: Any = None,
        batch_size: Any = None,
        outputs: Any = None,
        silent: bool = None,
    ):
        ...

    @sarus_method("shap.SHAP_SHAP_VALUES")
    def shap_values(self, X: Union[ndarray, DataFrame]):
        ...

    @sarus_method("shap.SHAP_CALL")
    def __call__(self, X: Union[ndarray, DataFrame]):
        ...


class Permutation(DataSpecWrapper[shap.explainers.Permutation]):
    @sarus_init("shap.SHAP_PERMUTATION")
    def __init__(
        self,
        model: Union[BaseEstimator, Tuple[Any, Any]],
        masker: Union[Tuple[Any, Any], np.ndarray, pd.DataFrame, spmatrix],
        output_names: Optional[List[str]] = None,
        link: Any = None,
        linearize_link: bool = True,
        feature_names: Optional[List[str]] = None,
        nsamples: int = 1000,
        feature_perturbation: Optional[str] = None,
    ):
        ...

    @sarus_method("shap.SHAP_EXPLAIN_ROW")
    def explain_row(
        self,
        row_args: Any = None,
        max_evals: Any = None,
        main_effects: Any = None,
        error_bounds: Any = None,
        batch_size: Any = None,
        outputs: Any = None,
        silent: bool = None,
    ):
        ...

    @sarus_method("shap.SHAP_PERMUTATION_SHAP_VALUES")
    def shap_values(
        self,
        X: Union[np.ndarray, pd.DataFrame],
        npermutations: Optional[int] = 10,
        main_effects: Optional[bool] = False,
        error_bounds: Optional[bool] = False,
        batch_evals: Optional[bool] = True,
        silent: Optional[bool] = False,
    ) -> Any:
        ...

    @sarus_method("shap.SHAP_CALL")
    def __call__(self, X: Union[ndarray, DataFrame]):
        ...


class Sampling(DataSpecWrapper[shap.explainers.Sampling]):
    @sarus_init("shap.SHAP_SAMPLING")
    def __init__(
        self,
        model: Union[BaseEstimator, Tuple[Any, Any]],
        data: Union[Tuple[Any, Any], np.ndarray, pd.DataFrame, spmatrix],
    ):
        ...

    @sarus_method("shap.SHAP_SAMPLING_ESTIMATE")
    def sampling_estimate(
        self,
        j: int,
        f: Callable,
        x: Union[pd.Series, pd.DataFrame, np.ndarray],
        X: Union[pd.Series, pd.DataFrame, np.ndarray],
        nsamples: Optional[int] = 10,
    ) -> Any:
        ...

    @sarus_method("shap.SHAP_RUN")
    def run(self):
        ...

    @sarus_method("shap.SHAP_ALLOCATE")
    def allocate(self):
        ...

    @sarus_method("shap.SHAP_SOLVE")
    def solve(self, fraction_evaluated: Any, dim: Any):
        ...

    @sarus_method("shap.SHAP_VARYING_GROUPS")
    def varying_groups(self, X: Any):
        ...

    @sarus_method("shap.SHAP_EXPLAIN")
    def explain(self, incoming_instance: Any):
        ...

    @sarus_method("shap.SHAP_ADD_SAMPLE")
    def addsample(self, x: np.array, m: np.array, w: float):
        ...

    @sarus_method("shap.SHAP_EXPLAIN_ROW")
    def explain_row(
        self,
        row_args: Any = None,
        max_evals: Any = None,
        main_effects: Any = None,
        error_bounds: Any = None,
        batch_size: Any = None,
        outputs: Any = None,
        silent: bool = None,
    ):
        ...

    @sarus_method("shap.SHAP_SHAP_VALUES")
    def shap_values(self, X: Union[ndarray, DataFrame]):
        ...

    @sarus_method("shap.SHAP_CALL")
    def __call__(self, X: Union[ndarray, DataFrame]):
        ...


class Exact(DataSpecWrapper[shap.explainers.Exact]):
    @sarus_init("shap.SHAP_EXACT")
    def __init__(
        self,
        model: Union[BaseEstimator, Tuple[Any, Any]],
        masker: Union[Tuple[Any, Any], np.ndarray, pd.DataFrame, spmatrix],
        link: Any = None,
        linearize_link: bool = True,
        feature_names: Optional[List[str]] = None,
    ):
        ...

    @sarus_method("shap.SHAP_EXPLAIN_ROW")
    def explain_row(
        self,
        row_args: Any = None,
        max_evals: Any = None,
        main_effects: Any = None,
        error_bounds: Any = None,
        batch_size: Any = None,
        outputs: Any = None,
        silent: bool = None,
    ):
        ...

    @sarus_method("shap.SHAP_SHAP_VALUES")
    def shap_values(self, X: Union[ndarray, DataFrame]):
        ...

    @sarus_method("shap.SHAP_CALL")
    def __call__(self, X: Union[ndarray, DataFrame]):
        ...


class Additive(DataSpecWrapper[shap.explainers.Additive]):
    @sarus_init("shap.SHAP_ADDITIVE")
    def __init__(
        self,
        model: Union[BaseEstimator, Tuple[Any, Any]],
        masker: Union[Tuple[Any, Any], np.ndarray, pd.DataFrame, spmatrix],
        link: Any = None,
        feature_names: Optional[List[str]] = None,
        linearize_link: Optional[bool] = True,
    ):
        ...

    @sarus_method("shap.SHAP_EXPLAIN_ROW")
    def explain_row(
        self,
        row_args: Any = None,
        max_evals: Any = None,
        main_effects: Any = None,
        error_bounds: Any = None,
        batch_size: Any = None,
        outputs: Any = None,
        silent: bool = None,
    ):
        ...

    @sarus_method("shap.SHAP_SHAP_VALUES")
    def shap_values(self, X: Union[ndarray, DataFrame]):
        ...

    @sarus_method("shap.SHAP_CALL")
    def __call__(self, X: Union[ndarray, DataFrame]):
        ...


register_ops()



"""Sarus Pandas package documentation."""
from sarus.utils import register_ops

from .explainers import *
from .Explanation import *
from .plots import *
from .maskers import *
from .utils import *

from shap import (
    bar_plot,
    summary_plot,
    decision_plot,
    multioutput_decision_plot,
    embedding_plot,
    force_plot,
    save_html,
    group_difference_plot,
    image_plot,
    monitoring_plot,
    partial_dependence_plot,
    dependence_plot,
    text_plot,
    waterfall_plot,
)

from .explainers import Additive as AdditiveExplainer
from .explainers import Explainer
from .explainers import GPUTree as GPUTreeExplainer
from .explainers import Kernel as KernelExplainer
from .explainers import Linear as LinearExplainer
from .explainers import Partition as PartitionExplainer
from .explainers import Permutation as PermutationExplainer
from .explainers import Sampling as SamplingExplainer
from .explainers import Tree as TreeExplainer

register_ops()



from __future__ import annotations

import sarus_data_spec.protobuf as sp

from sarus.dataspec_wrapper import DataSpecWrapper
from sarus.utils import register_ops, sarus_init, sarus_method

try:
    import sklearn.svm as svm
    from sklearn.svm import *
except ModuleNotFoundError:
    pass  # error message in sarus_data_spec.typing


class SVC(DataSpecWrapper[svm.SVC]):
    @sarus_init("sklearn.SK_SVC")
    def __init__(self, *args, **kwargs) -> None:
        ...

    @sarus_method("sklearn.SK_FIT", inplace=True)
    def fit(self, X, y, sample_weight=None):
        ...


register_ops()



from sarus.dataspec_wrapper import DataSpecWrapper
from sarus.utils import register_ops, sarus_init, sarus_method

try:
    import skopt.searchcv as skopt_searchcv
    from skopt.searchcv import *
except ModuleNotFoundError:
    pass  # error message in sarus_data_spec.typing


class BayesSearchCV(DataSpecWrapper[skopt_searchcv.BayesSearchCV]):
    @sarus_init("skopt.SKOPT_BAYES_SEARCH_CV")
    def __init__(
        self,
        estimator,
        search_spaces,
        optimizer_kwargs=None,
        n_iter=50,
        scoring=None,
        fit_params=None,
        n_jobs=1,
        n_points=1,
        iid=True,
        refit=True,
        cv=None,
        verbose=0,
        pre_dispatch='2*n_jobs',
        random_state=None,
        error_score='raise',
        return_train_score=False,
    ):
        ...

    @sarus_method("sklearn.SK_FIT", inplace=True)
    def fit(self, X, y=None):
        ...


register_ops()



from enum import Enum
from typing import (
    Any,
    Dict,
    Optional,
    Protocol,
    Type,
    TypeVar,
    Union,
    runtime_checkable,
)

import sarus_data_spec.context.typing as sct
import sarus_data_spec.typing as st
from requests import Session


class SyncPolicy(Enum):
    MANUAL = 0
    SEND_ON_INIT = 1
    SEND_ON_VALUE = 2


class DataSpecVariant(Enum):
    USER_DEFINED = "user_defined"
    SYNTHETIC = "synthetic"
    MOCK = "mock"
    ALTERNATIVE = "alternative"


T = TypeVar("T")


SPECIAL_WRAPPER_ATTRIBUTES = [
    "_alt_dataspec",
    "_dataspec",
    "_alt_policy",
    "_traced_transform",
    "_manager",
    "__sarus_idx__",
]

MOCK = "mock"
PYTHON_TYPE = "python_type"
ADMIN_DS = "admin_ds"
ATTRIBUTES_DS = "attributes_ds"


@runtime_checkable
class DataSpecWrapper(Protocol[T]):
    def python_type(self) -> Optional[str]:
        ...

    def dataspec(
        self, kind: DataSpecVariant = DataSpecVariant.USER_DEFINED
    ) -> st.DataSpec:
        ...

    def __sarus_eval__(
        self,
        target_epsilon: Optional[float] = None,
        verbose: Optional[int] = None,
    ) -> st.DataSpecValue:
        """Return value of synthetic variant."""
        ...

    def traced_transform(self) -> Optional[st.Transform]:
        ...

    def _set_traced_transform(
        self, traced_transform: Optional[st.Transform]
    ) -> None:
        ...

    def _set_dataspec(self, dataspec: st.DataSpec) -> None:
        ...

    def _set_variable(self, name: str, position: int):
        ...


@runtime_checkable
class TracedFunction(Protocol):
    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        ...

    def transform(self) -> st.Transform:
        ...


class DataSpecWrapperFactory(Protocol):
    def register(
        self,
        python_classname: str,
        sarus_wrapper_class: DataSpecWrapper,
    ) -> None:
        ...

    def create(self, dataspec: st.DataSpec) -> Union[DataSpecWrapper, Any]:
        """Create a wrapper class from a DataSpec.

        If the dataspec's python value is not managed by the SDK, returns an
        unwrapped Python object.
        """

    def from_value(self, value: Any) -> DataSpecWrapper:
        ...

    def registry(self) -> Dict[str, Type[DataSpecWrapper]]:
        ...


class Client:
    def url(self) -> str:
        """Return the URL of the Sarus server."""

    def session(self) -> Session:
        """Return the connection to the server."""

    def context(self) -> sct.Context:
        """Return the client's context."""



import sarus_data_spec.status as stt
from sarus_data_spec import typing as st
from sarus_data_spec.manager.computations.base import status_error_policy
from sarus_data_spec.manager.computations.remote.base import (
    RemoteComputation as BaseRemote,
)
from sarus_data_spec.manager.computations.remote.base import T

from .dataspec_api import launch_dataspec


class RemoteComputation(BaseRemote[T]):
    async def error(
        self,
        dataspec: st.DataSpec,
    ) -> st.Status:
        """For Remote computations, when mthere is an error status
        to be relaunched, a new api call is done"""
        status = self.status(dataspec)
        if status is not None:
            stage = status.task(self.task_name)
            assert stage
            should_clear = status_error_policy(stage=stage)
            if should_clear:
                launch_dataspec(
                    self.computing_manager().client(), dataspec.uuid()
                )
                return await self.complete_task(dataspec=dataspec)
            raise stt.DataSpecErrorStatus(
                (
                    stage.properties()["relaunch"] == str(True),
                    stage.properties()["message"],
                )
            )
        return await self.complete_task(dataspec=dataspec)



from __future__ import annotations

import logging
import typing as t

import pandas as pd

from sarus.dataspec_wrapper import (
    IGNORE_WARNING,
    DataSpecVariant,
    DataSpecWrapper,
)
from sarus.typing import SPECIAL_WRAPPER_ATTRIBUTES
from sarus.utils import (
    create_lambda_op,
    create_method,
    register_ops,
    sarus_init,
    sarus_method,
    sarus_property,
)

logger = logging.getLogger(__name__)
from shap import Explainer
from typing import (
    Any,
    Callable,
    Dict,
    Hashable,
    Iterable,
    List,
    Literal,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
)

from pandas._typing import Axes
from typing import Any
import shap
import numpy.typing as npt
from scipy.sparse import spmatrix


class Explanation(DataSpecWrapper[shap.Explanation]):
    @sarus_init("shap.SHAP_EXPLANATION")
    def __init__(
        self,
        values: Union[npt.ArrayLike, spmatrix],
        base_values: Optional[Union[npt.ArrayLike, spmatrix]] = None,
        data: Optional[Union[npt.ArrayLike, spmatrix]] = None,
        display_data: Optional[Dict[str, npt.ArrayLike]] = None,
        instance_names: Optional[List[str]] = None,
        feature_names: Optional[List[str]] = None,
        output_names: Optional[List[str]] = None,
        output_indexes: Optional[List[int]] = None,
        lower_bounds: Optional[Union[npt.ArrayLike, spmatrix]] = None,
        upper_bounds: Optional[Union[npt.ArrayLike, spmatrix]] = None,
        error_std: Optional[Union[npt.ArrayLike, spmatrix]] = None,
        main_effects: Optional[Union[npt.ArrayLike, spmatrix]] = None,
        hierarchical_values: Optional[Union[npt.ArrayLike, spmatrix]] = None,
        clustering: Optional[Union[npt.ArrayLike, spmatrix]] = None,
        compute_time: Optional[float] = None,
    ):
        ...

    @sarus_property("shap.SHAP_VALUES")
    def values(self):
        ...

    @sarus_property("shap.SHAP_SUM")
    def sum(self):
        ...



import io

import pyarrow as pa
import pyarrow.parquet as pq
import requests

from sarus.typing import Client


def fetch_synthetic(client: Client, id: str) -> pa.Table:
    """Fetch synthetic data for a Dataset."""
    resp = client.session().get(
        f"{client.url()}/synthetic_data/{id}",
        stream=True,
        params={
            "textual_categories": True,
            "rows_number": None,
        },
    )
    if resp.status_code > 200:
        raise Exception(
            f"Error while retrieving synthetic data. "
            f"Gateway answer was: \n{resp}"
        )

    synthetic_table = pq.ParquetFile(io.BytesIO(resp.content)).read()

    return synthetic_table



import typing as t

import pyarrow as pa
import sarus_data_spec.status as stt
from sarus_data_spec import typing as st
from sarus_data_spec.constants import ARROW_TASK
from sarus_data_spec.manager.computations.base import (
    ErrorCatchingAsyncIterator,
)

from sarus.manager.base_remote import RemoteComputation

from .dataspec_api import dataset_result, dataspec_status, launch_dataspec


class ToArrowComputationOnServer(
    RemoteComputation[t.AsyncIterator[pa.RecordBatch]]
):
    """ToArrowComputation on the Sarus server through the REST API."""

    task_name = ARROW_TASK

    def launch_task(self, dataspec: st.DataSpec) -> None:
        """Launch the computation of a Dataset on the server."""
        status = self.status(dataspec)
        if status is None:
            launch_dataspec(self.computing_manager().client(), dataspec.uuid())

    async def result_from_stage_properties(
        self,
        dataspec: st.DataSpec,
        properties: t.Mapping[str, str],
        **kwargs: t.Any,
    ) -> t.AsyncIterator[pa.RecordBatch]:
        """Return the task result, launch the computation if needed."""
        batch_size = kwargs["batch_size"]
        ait = dataset_result(
            self.computing_manager().client(), dataspec.uuid(), batch_size
        )
        return ErrorCatchingAsyncIterator(ait, dataspec, self)

    def status(self, dataspec: st.DataSpec) -> t.Optional[st.Status]:
        """In this case, the computation retrieves the status from the server
        via an API call, a local status is created but not stored in order
        to use it just for the polling"""

        status_proto = dataspec_status(
            client=self.computing_manager().client(),
            uuid=dataspec.uuid(),
            task_names=[self.task_name],
        )
        if status_proto is None:
            return None
        else:
            return stt.Status(protobuf=status_proto, store=False)



from __future__ import annotations

import inspect
import typing as t
from functools import wraps

import sarus_data_spec.typing as st
from sarus_data_spec.context import global_context

from sarus.context.typing import LocalSDKContext
from sarus.dataspec_wrapper import DataSpecWrapper
from sarus.utils import convert_container


def trace(
    python_callable: t.Callable[..., t.Any]
) -> t.Callable[..., TracedFunction]:
    """Serialize a python callable into a Transform.

    The transform can then be safely executed on the server.

    The syntax would be `traced_function = trace(preprocess)(example_data)`
    similar to Jax's jit.
    """

    @wraps(python_callable)
    def tracing_fn(
        *args: DataSpecWrapper, **kwargs: DataSpecWrapper
    ) -> TracedFunction:
        # Make tracing copies of the input arguments
        tracing_args = []
        tracing_kwargs = {}
        args_names = [
            param for param in inspect.signature(python_callable).parameters
        ]
        context: LocalSDKContext = global_context()
        for i, arg in enumerate(args):
            if not isinstance(arg, DataSpecWrapper):
                arg = context.wrapper_factory().from_value(arg)
            tracing_arg = type(arg).from_dataspec(arg._dataspec)
            tracing_arg._set_variable(name=args_names[i], position=i)
            tracing_args.append(tracing_arg)

        for name, arg in kwargs.items():
            if not isinstance(arg, DataSpecWrapper):
                arg = context.wrapper_factory().from_value(arg)
            tracing_arg = type(arg).from_dataspec(arg._dataspec)
            tracing_arg._set_variable(
                name=name, position=args_names.index(name)
            )
            tracing_kwargs[name] = tracing_arg

        # Call the user-defined callable with the tracing arguments
        traced_result = python_callable(*tracing_args, **tracing_kwargs)
        traced_result = convert_container(traced_result)

        # Check that the result is indeed traced
        if not isinstance(traced_result, DataSpecWrapper):
            raise TypeError(
                "Error while serializing: the traced "
                "result is not a DataSpecWrapper."
            )

        traced_transform = traced_result.traced_transform()
        if traced_transform is None:
            raise ValueError(
                "Error while serializing: the traced transform is None."
            )

        return TracedFunction(traced_transform)

    return tracing_fn


class TracedFunction:
    def __init__(self, transform: st.Transform) -> None:
        self._transform = transform
        self._callable = transform.composed_callable()

    def __call__(self, *args: t.Any, **kwargs: t.Any) -> t.Any:
        return self._callable(*args, **kwargs)

    def transform(self) -> st.Transform:
        return self._transform

    def dot(self) -> str:
        return self._transform.dot()



import io
import json
import logging
import pickle as pkl
import time
import typing as t
from http import HTTPStatus

import pyarrow as pa
import pyarrow.parquet as pq
import sarus_data_spec.protobuf as sp
import sarus_data_spec.typing as st
from requests import Response
from sarus_data_spec.constants import (
    BEST_ALTERNATIVE,
    CONSTRAINT_KIND,
    PRIVACY_LIMIT,
    SCHEMA_TASK,
)
from sarus_data_spec.dataspec_validator.typing import DataspecPrivacyPolicy
from sarus_data_spec.manager.async_utils import async_iter
from sarus_data_spec.protobuf.utilities import dict_deserialize, dict_serialize

from sarus.typing import Client

logger = logging.getLogger(__name__)

CHUNK_SIZE = 1024 * 1024


def raise_response(resp: Response) -> None:
    """Raise exception with message encapsulated in the response JSON data."""
    if resp.status_code >= HTTPStatus.INTERNAL_SERVER_ERROR:
        resp.raise_for_status()
    if resp.status_code >= HTTPStatus.BAD_REQUEST:
        try:
            message = resp.json().get("message")
        except json.JSONDecodeError as e:
            resp.raise_for_status()
        if message is not None:
            raise ValueError(f"Server - {message}")
    resp.raise_for_status()


def get_dataspec(
    client: Client, uuid: str
) -> t.Tuple[t.Optional[st.DataSpec], float]:
    """Fetch a single Dataspec from the server."""
    start = time.perf_counter()
    resp: Response = client.session().get(
        f"{client.url()}/dataspecs/{uuid}",
    )
    end = time.perf_counter()
    logger.debug(
        f"GET /dataspecs/{uuid} {resp.status_code} ({end-start:.2f}s)"
    )
    if resp.status_code == HTTPStatus.NOT_FOUND:
        return None

    raise_response(resp)

    proto = dict_deserialize(resp.json()["dataspec"])
    return client.context().factory().create(proto)


def get_epsilon(
    client: Client, uuid: str
) -> t.Tuple[t.Optional[st.DataSpec], float]:
    """Fetch a single Dataspec from the server and the consumed epsilon."""
    resp: Response = client.session().get(
        f"{client.url()}/dataspecs/{uuid}/epsilon",
    )

    if resp.status_code == HTTPStatus.NOT_FOUND:
        return None, 0.0

    raise_response(resp)

    epsilon = resp.json()["epsilon"]
    return epsilon


def pull_dataspec_graph(client: Client, uuid: str) -> None:
    """Fetch a dataspec's computation graph and store it."""
    start = time.perf_counter()
    resp: Response = client.session().get(
        f"{client.url()}/dataspecs/{uuid}/graph"
    )
    end = time.perf_counter()
    logger.debug(
        f"GET /dataspecs/{uuid}/graph {resp.status_code} ({end-start:.2f}s)"
    )
    raise_response(resp)

    protos = [dict_deserialize(msg) for msg in resp.json()]
    referrables = [
        client.context().factory().create(proto, store=False)
        for proto in protos
    ]
    client.context().storage().batch_store(referrables)


def pull_dataspec_schema(client: Client, uuid: str) -> None:
    """Fetch a dataspec's computation graph and store it."""
    start = time.perf_counter()
    resp: Response = client.session().get(
        f"{client.url()}/dataspecs/{uuid}/schema"
    )
    end = time.perf_counter()
    logger.debug(
        f"GET /dataspecs/{uuid}/schema {resp.status_code} ({end-start:.2f}s)"
    )
    raise_response(resp)

    dataspec = client.context().storage().referrable(uuid)
    proto_schema = dict_deserialize(resp.json()["schema"])
    schema = client.context().factory().create(proto_schema)
    client.context().manager().copy_status_from_server(
        dataspec, task_names=[SCHEMA_TASK]
    )
    return schema


def push_dataspec_graph(client: Client, graph: t.List[st.Referrable]) -> None:
    """Push a list of referrables to the server."""
    start = time.perf_counter()
    resp: Response = client.session().post(
        f"{client.url()}/dataspecs/graph",
        json=[dict_serialize(ref.protobuf()) for ref in graph],
    )
    end = time.perf_counter()
    logger.debug(
        f"POST /dataspecs/graph {resp.status_code} ({end-start:.2f}s)"
    )
    raise_response(resp)


def compile_dataspec(
    client: Client,
    uuid: str,
    constraint_kind: t.Optional[st.ConstraintKind] = None,
    privacy_limit: t.Optional[st.PrivacyLimit] = None,
) -> t.Tuple[str, t.Optional[DataspecPrivacyPolicy]]:
    """Compile the dataspec to abide privacy constraints.

    Return the compiled dataspec's UUID and the optional privacy policy name.
    """
    kind_name = constraint_kind.name if constraint_kind else BEST_ALTERNATIVE
    payload = {CONSTRAINT_KIND: kind_name}
    if privacy_limit is not None:
        payload[PRIVACY_LIMIT] = privacy_limit.delta_epsilon_dict()

    start = time.perf_counter()
    resp: Response = client.session().post(
        f"{client.url()}/dataspecs/{uuid}/compile",
        json=payload,
    )
    end = time.perf_counter()
    logger.debug(
        f"POST /dataspecs/{uuid}/compile {resp.status_code} ({end-start:.2f}s)"
    )
    raise_response(resp)

    pp_value = resp.json()["privacy_policy"]
    privacy_policy = DataspecPrivacyPolicy(pp_value) if pp_value else None

    return resp.json()["uuid"], privacy_policy


def launch_dataspec(client: Client, uuid: str) -> None:
    """Launch a Dataspec's computation on the server."""
    start = time.perf_counter()
    resp: Response = client.session().post(
        f"{client.url()}/dataspecs/{uuid}/launch",
    )
    end = time.perf_counter()
    logger.debug(
        f"POST /dataspecs/{uuid}/launch {resp.status_code} ({end-start:.2f}s)"
    )
    raise_response(resp)


def dataspec_status(
    client: Client, uuid: str, task_names: t.List[str]
) -> t.Optional[sp.Status]:
    """Get the dataspec's status on the server."""
    if type(task_names) not in [set, list, tuple]:
        raise TypeError("task_names should be a list of strings.")

    start = time.perf_counter()
    resp: Response = client.session().get(
        f"{client.url()}/dataspecs/{uuid}/status",
        params={"task_names": list(task_names)},
    )
    end = time.perf_counter()
    logger.debug(
        f"GET /dataspecs/{uuid}/status {resp.status_code} ({end-start:.2f}s)"
    )
    raise_response(resp)

    status_proto = resp.json().get("status")
    if status_proto is None:
        return None
    else:
        return dict_deserialize(status_proto)


def pull_dataspec_status_graph(
    client: Client, uuid: str, task_names: t.List[str]
) -> t.List[sp.Status]:
    """Fetch the server statuses of the computation graph's dataspecs."""
    if not type(task_names) in [list, set, tuple]:
        raise TypeError("task_names should be a list of strings.")

    start = time.perf_counter()
    resp: Response = client.session().get(
        f"{client.url()}/dataspecs/{uuid}/graph/statuses",
        params={"task_names": list(task_names)},
    )
    end = time.perf_counter()
    logger.debug(
        f"GET /dataspecs/{uuid}/graph/statuses {resp.status_code} ({end-start:.2f}s)"
    )
    raise_response(resp)

    return [dict_deserialize(msg) for msg in resp.json()]


def dataspec_result_response(
    client: Client, uuid: str, batch_size: t.Optional[int] = None
) -> Response:
    """Return the response result from the server.

    The response holds the dataspec's value and is read in the computation.
    """
    start = time.perf_counter()
    resp: Response = client.session().get(
        f"{client.url()}/dataspecs/{uuid}/result",
        params={"batch_size": batch_size},
        stream=True,
    )
    end = time.perf_counter()
    logger.debug(
        f"GET /dataspecs/{uuid}/result {resp.status_code} {resp.headers.get('Content-Type')} ({end-start:.2f}s)"
    )
    raise_response(resp)
    return resp


def dataset_result(
    client: Client, uuid: str, batch_size: int
) -> t.AsyncIterator[pa.RecordBatch]:
    """Return the dataset's value as a RecordBatch async iterator."""
    resp = dataspec_result_response(client, uuid, batch_size)
    if resp.headers.get("Content-Type") == "application/parquet":
        # Recieving Parquet file
        buffer = io.BytesIO()
        for data in resp.iter_content(CHUNK_SIZE):
            buffer.write(data)
        buffer.seek(0)
        return async_iter(
            pq.read_table(buffer).to_batches(max_chunksize=batch_size)
        )
    else:
        # Recieving serialized streamed record batches
        async def arrow_iterator_from_response():
            with pa.ipc.open_stream(resp.raw) as reader:
                for batch in reader:
                    yield batch

        return arrow_iterator_from_response()


def scalar_result(client: Client, uuid: str) -> t.Any:
    """Return the scalar's value."""
    resp = dataspec_result_response(client, uuid)
    return pkl.loads(resp.content)



import logging
import pickle as pkl
import time
import traceback
import typing as t

import sarus_data_spec.protobuf as sp
import sarus_data_spec.status as stt
from sarus_data_spec import typing as st
from sarus_data_spec.constants import SCALAR_TASK, ScalarCaching
from sarus_data_spec.manager.base import Base
from sarus_data_spec.manager.computations.local.base import LocalComputation
from sarus_data_spec.scalar import Scalar

from sarus.manager.cache_scalar_local import CacheScalarComputation
from sarus.manager.value_remote import ValueComputationOnServer

logger = logging.getLogger(__name__)


class ValueComputation(LocalComputation[t.Any]):
    """Class responsible for handling the computation
    of scalars."""

    task_name = SCALAR_TASK

    def __init__(
        self,
        computing_manager: Base,
        cache_scalar_computation: CacheScalarComputation,
        remote_scalar: ValueComputationOnServer,
    ) -> None:
        super().__init__(computing_manager)
        self.cache_scalar_computation = cache_scalar_computation
        self.remote_scalar = remote_scalar

    async def result_from_stage_properties(
        self,
        dataspec: st.DataSpec,
        properties: t.Mapping[str, str],
        **kwargs: t.Any,
    ) -> t.Any:
        if self.computing_manager().is_cached(dataspec):
            (
                cache_type,
                cache,
            ) = await self.cache_scalar_computation.task_result(dataspec)
            try:
                if cache_type == ScalarCaching.PICKLE.value:
                    with open(cache, "rb") as f:
                        data = pkl.load(f)

                else:
                    data = sp.python_proto_factory(cache, cache_type)
            except Exception as e:
                stt.error(
                    dataspec=dataspec,
                    manager=self.computing_manager(),
                    task=self.task_name,
                    properties={
                        "message": traceback.format_exc(),
                        'relaunch': str(True),
                    },
                )
                stt.error(
                    dataspec=dataspec,
                    manager=self.computing_manager(),
                    task=self.cache_scalar_computation.task_name,
                    properties={
                        "message": traceback.format_exc(),
                        'relaunch': str(True),
                    },
                )
                raise stt.DataSpecErrorStatus(
                    (True, traceback.format_exc())
                ) from e
            else:
                return data

        if self.computing_manager().is_computation_remote(dataspec):
            status = self.remote_scalar.status(dataspec)
            return await self.remote_scalar.result_from_stage_properties(
                dataspec, properties=status.task(self.task_name).properties()
            )

        return await self.computing_manager().async_value_op(
            scalar=t.cast(Scalar, dataspec)
        )

    async def prepare(self, dataspec: st.DataSpec) -> None:
        try:
            logger.debug(f'STARTED LOCAL SCALAR {dataspec.uuid()}')
            start = time.perf_counter()
            if self.computing_manager().is_cached(dataspec):
                await self.cache_scalar_computation.task_result(dataspec)

            elif self.computing_manager().is_computation_remote(dataspec):
                await self.remote_scalar.complete_task(dataspec)
            else:
                await self.computing_manager().async_prepare_parents(dataspec)
        except stt.DataSpecErrorStatus as exception:
            stt.error(
                dataspec=dataspec,
                manager=self.computing_manager(),
                task=self.task_name,
                properties={
                    "message": traceback.format_exc(),
                    'relaunch': str(exception.relaunch),
                },
            )
            raise
        except Exception:
            stt.error(
                dataspec=dataspec,
                manager=self.computing_manager(),
                task=self.task_name,
                properties={
                    "message": traceback.format_exc(),
                    'relaunch': str(False),
                },
            )
            raise stt.DataSpecErrorStatus((False, traceback.format_exc()))
        else:
            end = time.perf_counter()
            logging.debug(
                f'FINISHED LOCAL SCALAR {dataspec.uuid()} ({end-start:.2f}s)'
            )
            stt.ready(
                dataspec=dataspec,
                manager=self.computing_manager(),
                task=self.task_name,
            )



import typing as t

import sarus_data_spec.typing as st
from sarus_data_spec.dataspec_validator.typing import DataspecPrivacyPolicy
from sarus_data_spec.manager.base import Computation
from sarus_data_spec.manager.typing import Manager

from sarus.typing import Client


class SDKManager(Manager, t.Protocol):
    """The Manager of the SDK running on the client side.

    This Manager has two additional functionalities compared to the
    DelegatingManager manager.

    First, it manages the relationship with the remote server using the API
    endpoints.

    Second, this Manager defines a MOCK version for every DataSpec. The MOCK is
    defined as a smaller version of a DataSpec. In practice, it is a sample of
    SYNTHETIC at the source and MOCKs of transformed DataSpecs are the
    transforms of the MOCKs.

    The MOCK is created and its value computed in the `infer_output_type`
    method. This serves two purposes. First, it provides immediate feedback to
    the user in case of erroneous computation. Second, it allows identifying the
    MOCK's value Python type which is then used by the SDK to instantiate the
    correct DataSpecWrapper type (e.g. instantiate a sarus.pandas.DataFrame if
    the value is a pandas.DataFrame).
    """

    def python_type(self, dataspec: st.DataSpec) -> t.Optional[str]:
        """Return the Python class name of a DataSpec.

        This is used to instantiate the correct DataSpecWrapper class.
        """

    def client(self) -> Client:
        """Return the sarus.Client object used to make API calls."""

    def launch(self, dataspec: st.DataSpec) -> None:
        """Launch a Dataspec's computation on the server."""

    def compile(
        self, dataspec: st.DataSpec, target_epsilon: t.Optional[float] = None
    ) -> t.Tuple[st.DataSpec, DataspecPrivacyPolicy]:
        """Compile an alternative Dataspec on the server."""

    def push(self, dataspec: st.DataSpec) -> None:
        """Push a Dataspec's computation graph on the server."""

    def dataspec_computation(self, dataspec: st.DataSpec) -> Computation:
        """Return the Computation for getting the dataspec's value. If sql is
        is true, the sqlcomputation will be returned.
        """

    def set_admin_ds(
        self, source_ds: st.DataSpec, admin_ds: t.Dict[str, t.Any]
    ) -> None:
        """Attach the admin_ds to a source dataspec in a status."""

    def default_delta(self, dataspec: st.DataSpec) -> t.Optional[float]:
        """Get the default delta of a dataspec."""

    def consumed_epsilon(self, dataspec: st.DataSpec) -> float:
        """Get the consumed epsilon from the server."""



import pandas as pd

from sarus.dataspec_wrapper import DataSpecWrapper
from sarus.utils import register_ops


class PrettyDict(DataSpecWrapper[pd.io.formats.printing.PrettyDict]):
    ...


register_ops()



from enum import Enum
from typing import Protocol, runtime_checkable

from sarus_data_spec.context.typing import Context

from sarus.typing import DataSpecWrapperFactory


@runtime_checkable
class LocalSDKContext(Context, Protocol):
    def wrapper_factory(self) -> DataSpecWrapperFactory:
        ...

    def verbose(self) -> int:
        ...
